(*
 * SNU 4190.310 Programming Languages (Fall 2006)
 *
 * Vanilla M: interpreter with dynamic type checking
 *)

structure M_Vanilla : M_Runner = struct
    open M

    // domains
    type loc = int
    type value = Num of int
               | String of string
               | Bool of bool
               | Loc of loc
               | Pair of value * value
               | Closure of closure
    and closure = fexpr * env
    and fexpr = Fun of id * exp
              | RecFun of id * id * exp
    and env = id -> value
    type memory = int * (loc -> value)

    // notations (see 5 page in M.pdf)
    (* f @+ (x, v)              f[x |-> v]
     * store M (l, v)           M[l |-> v]
     * fetch M l                M(l)
     *)
    fun (@+) f (x, v) = (fn y => if y = x then v else f y)
    fun store (l, m) p = (l, m @+ p)        
    fun fetch (_, m) l = m l                
    fun malloc (l, m) = (l, (l+1, m))

    // auxiliary functions
    fun error msg = raise RuntimeError msg
    val getClosure = fn Closure c => c | _ => error "not a function"
    val getNum = fn Num n => n | _ => error "not a number value"
    val getBool = fn Bool b => b | _ => error "not a boolean value"
    val getLoc = fn Loc l => l | _ => error "not a location value"
    val getPair = fn Pair p => p | _ => error "not a pair"
    val op2fn =
        fn ADD => (fn (v1,v2) => Num (getNum v1 + getNum v2))
         | SUB => (fn (v1,v2) => Num (getNum v1 - getNum v2))
         | AND => (fn (v1,v2) => Bool (getBool v1 andalso getBool v2))
         | OR =>  (fn (v1,v2) => Bool (getBool v1 orelse getBool v2))
         | EQ =>  (* TODO: implementation *)
            raise RuntimeError "not implemented" 

    val rec printValue =
        fn Num n => print_int n; print_newline()
         | Bool b => print_endline (if b then "true" else "false")
         | String s => print_endline s
         | _ => error "unprintable"

    fun eval env mem exp = case exp
        of CONST c =>
            (case c of S s => String s | N n => Num n | B b => Bool b, mem)
         | VAR x => (env x, mem)
         | FN (x, e) => (Closure (Fun (x, e), env), mem)
         | APP (e1, e2) =>
            let val (v1, m') = eval env mem e1
                val (c, env') = getClosure v1
                val (v2, m'') = eval env m' e2
            in  case c of Fun (x, e) => eval (env' @+ (x,v2)) m'' e
                        | RecFun (f, x, e) => (* TODO: implementation *)
                            raise RuntimeError "not implemented"
            end
         | LET (NREC (x, e1), e2) =>
            let val (v1, m') = eval env mem e1
            in  eval (env @+ (x,v1)) m' e2
            end
         | LET (REC (f, e1), e2) =>
            let val (v1, m') = eval env mem e1
                val (c, env') = getClosure v1
            in  
                (* TODO: implementation *)
                raise RuntimeError "not implemented"
            end
         | IF (e1, e2, e3) =>
            let val (v1, m') = eval env mem e1
            in  eval env m' (if getBool v1 then e2 else e3)
            end
         | BOP (op, e1, e2) =>
            let val (v1, m') = eval env mem e1
                val (v2, m'') = eval env m' e2
            in  ((op2fn op) (v1,v2), m'')
            end
         | READ =>
            let val n = read_int () handle _ => error "read error"
            in  (Num n, mem)
            end
         | WRITE e =>
            let val (v, m') = eval env mem e
            in  printValue v; (v, m')
            end
         | MALLOC e =>
            let val (v, m') = eval env mem e
                val (l, m'') = malloc m'
            in  (Loc l, store m'' (l,v))
            end
         | _ => (* TODO: implementation *)
            raise RuntimeError "not implemented"

    val emptyEnv = (fn x => error ("unknown id: " ^ x))
    val emptyMem = (0, fn l => error ("unknown loc: " ^ string_of_int l))

    fun run exp = ignore (eval emptyEnv emptyMem exp)
end
